@@ -171,7 +171,8 @@ Otherwise, prompt the user to enter a system message."
   (declare-function llm-ollama-chat-model "ext:llm-ollama")
   (declare-function llm-ollama-default-chat-temperature "ext:llm-ollama")
   (declare-function llm-ollama-default-chat-non-standard-params "ext:llm-ollama")
-  (when (llm-ollama-p provider)
+  (when (and (fboundp 'llm-ollama-p)
+             (llm-ollama-p provider))
     (setq ellama-transient-ollama-model-name (llm-ollama-chat-model provider))
     (setq ellama-transient-temperature (or (llm-ollama-default-chat-temperature provider) 0.7))
     (setq ellama-transient-host (llm-ollama-host provider))
